from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    _TypedDict,
)

try:
    from typing import TypeVarTuple, Unpack
except ImportError:
    from typing_extensions import TypeVarTuple, Unpack

from numpy import dtype, ndarray

from .python_serde import PythonSerde

PythonType = TypeVar("PythonType")
KeyPythonType = TypeVar("KeyPythonType")
ValuePythonType = TypeVar("ValuePythonType")
ItemsPythonType = TypeVar("ItemsPythonType")
PythonTypes = TypeVarTuple("PythonTypes")

class DynPyAnySerde(Generic[PythonType]): ...

class DynPyAnySerdeFactory:
    @staticmethod
    def bool_serde() -> DynPyAnySerde[bool]: ...
    @staticmethod
    def bytes_serde() -> DynPyAnySerde[bytes]: ...
    @staticmethod
    def complex_serde() -> DynPyAnySerde[complex]: ...
    @staticmethod
    def dict_serde(
        key_serde_option: Optional[DynPyAnySerde[KeyPythonType]],
        value_serde_option: Optional[DynPyAnySerde[ValuePythonType]],
    ) -> DynPyAnySerde[Dict[KeyPythonType, ValuePythonType]]: ...
    @staticmethod
    def dynamic_serde() -> DynPyAnySerde[Any]: ...
    @staticmethod
    def float_serde() -> DynPyAnySerde[float]: ...
    @staticmethod
    def int_serde() -> DynPyAnySerde[int]: ...
    @staticmethod
    def list_serde(
        items_serde_option: Optional[DynPyAnySerde[ItemsPythonType]],
    ) -> DynPyAnySerde[List[ItemsPythonType]]: ...
    @staticmethod
    def numpy_dynamic_shape_serde(py_dtype: dtype) -> DynPyAnySerde[ndarray]: ...
    @staticmethod
    def option_serde(
        value_serde_option: Optional[DynPyAnySerde[PythonType]],
    ) -> DynPyAnySerde[Optional[PythonType]]: ...
    @staticmethod
    def pickle_serde() -> DynPyAnySerde[Any]: ...
    @staticmethod
    def python_serde_serde(
        python_serde: PythonSerde[PythonType],
    ) -> DynPyAnySerde[PythonType]: ...
    @staticmethod
    def set_serde(
        items_serde_option: Optional[DynPyAnySerde[ItemsPythonType]],
    ) -> DynPyAnySerde[Set[ItemsPythonType]]: ...
    @staticmethod
    def string_serde() -> DynPyAnySerde[str]: ...
    @staticmethod
    def tuple_serde(
        item_serdes: List[Optional[DynPyAnySerde]],
    ) -> DynPyAnySerde[Tuple[Unpack[PythonTypes]]]: ...
    @staticmethod
    def typed_dict_serde(
        serde_dict: Dict[str, Optional[DynPyAnySerde]]
    ) -> DynPyAnySerde[_TypedDict]: ...
    @staticmethod
    def union_serde(
        serde_options: List[Optional[DynPyAnySerde]],
        serde_choice_fn: Callable[[Union[Unpack[PythonTypes]]], int],
    ) -> DynPyAnySerde[Union[Unpack[PythonTypes]]]: ...



from __future__ import annotations

from typing import Callable, Dict, List, Optional, Type, TypeVar, Union

try:
    from typing import TypeVarTuple, Unpack
except ImportError:
    from typing_extensions import TypeVarTuple, Unpack

import numpy as np

from .pyany_serde import DynPyAnySerde, DynPyAnySerdeFactory
from .python_serde import PythonSerde

T = TypeVar("T")
KeyT = TypeVar("KeyT")
ValueT = TypeVar("ValueT")
ItemsT = TypeVar("ItemsT")
Ts = TypeVarTuple("Ts")


def bool_serde():
    return DynPyAnySerdeFactory.bool_serde()


def bytes_serde():
    return DynPyAnySerdeFactory.bytes_serde()


def complex_serde():
    return DynPyAnySerdeFactory.complex_serde()


def dict_serde(
    key_serde: Optional[DynPyAnySerde[KeyT]],
    value_serde: Optional[DynPyAnySerde[ValueT]],
):
    return DynPyAnySerdeFactory.dict_serde(key_serde, value_serde)


def dynamic_serde():
    return DynPyAnySerdeFactory.dynamic_serde()


def float_serde():
    return DynPyAnySerdeFactory.float_serde()


def int_serde():
    return DynPyAnySerdeFactory.int_serde()


def list_serde(items_serde: Optional[DynPyAnySerde[ItemsT]]):
    return DynPyAnySerdeFactory.list_serde(items_serde)


def numpy_serde(dtype: Type[np._DTypeScalar_co]):
    return DynPyAnySerdeFactory.numpy_dynamic_shape_serde(np.dtype(dtype))


def option_serde(value_serde: Optional[DynPyAnySerde[T]]):
    return DynPyAnySerdeFactory.option_serde(value_serde)


def pickle_serde():
    return DynPyAnySerdeFactory.pickle_serde()


def python_serde_serde(python_serde: PythonSerde[T]):
    return DynPyAnySerdeFactory.python_serde_serde(python_serde)


def set_serde(items_serde: Optional[DynPyAnySerde[ItemsT]]):
    return DynPyAnySerdeFactory.set_serde(items_serde)


def string_serde():
    return DynPyAnySerdeFactory.string_serde()


def tuple_serde(*item_serdes: List[Optional[DynPyAnySerde]]):
    return DynPyAnySerdeFactory.tuple_serde(item_serdes)


def typed_dict_serde(serde_dict: Dict[str, Optional[DynPyAnySerde]]):
    return DynPyAnySerdeFactory.typed_dict_serde(serde_dict)


def union_serde(
    serde_options: List[Optional[DynPyAnySerde]],
    serde_choice_fn: Callable[[Union[Unpack[Ts]]], int],
):
    return DynPyAnySerdeFactory.union_serde(serde_options, serde_choice_fn)



from .pyany_serde import DynPyAnySerde, DynPyAnySerdeFactory
from .pyany_serde_factory import (
    bool_serde,
    bytes_serde,
    complex_serde,
    dict_serde,
    dynamic_serde,
    float_serde,
    int_serde,
    list_serde,
    numpy_serde,
    option_serde,
    pickle_serde,
    python_serde_serde,
    set_serde,
    string_serde,
    tuple_serde,
    typed_dict_serde,
    union_serde,
)
from .python_serde import PythonSerde



from abc import abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")


class PythonSerde(Generic[T]):
    @abstractmethod
    def to_bytes(self, obj: T) -> bytes:
        """
        Function to convert obj to bytes, for passing between batched agent and the agent manager.
        :return: bytes b such that from_bytes(b) == obj.
        """
        raise NotImplementedError

    @abstractmethod
    def from_bytes(self, byts: bytes) -> T:
        """
        Function to convert bytes to T, for passing between batched agent and the agent manager.
        :return: T obj such that from_bytes(to_bytes(obj)) == obj.
        """
        raise NotImplementedError
